/*
   Evaluate fully parenthesized infix expressions.
   This algorithm uses two stacks, one for operators
   and one for operands.

   This is the algorithm from page 329 of the textbook.
*/
import java.util.StringTokenizer;
import java.util.Stack;

public class EvaluateInfixTwoStacks
{

   public static int evaluateInfix(String infix)
   {
      Stack<String> operators = new Stack<String>();
      Stack<Integer> operands  = new Stack<Integer>();
      StringTokenizer st = new StringTokenizer(infix);

      while (st.hasMoreTokens())
      {
         String token = st.nextToken();
         if ( token.equals("(") )
         {
            /* do nothing */
         }
         else if ( token.equals("+")
                || token.equals("-")
                || token.equals("^")
                || token.equals("*")
                || token.equals("/")
                || token.equals("%") )
         {
            operators.push( token );
         }
         else if ( token.equals(")") )
         {
            String op = operators.pop();
            int y = operands.pop();
            int x = operands.pop();
            if ( op.equals("+") )
            {
               operands.push( x + y );
            }
            else if ( op.equals("-") )
            {
               operands.push( x - y );
            }
            else if ( op.equals("*") )
            {
               operands.push( x * y );
            }
            else if ( op.equals("/") )
            {
               operands.push( x / y );
            }
            else if ( op.equals("%") )
            {
               operands.push( x % y );
            }
            else if ( op.equals("^") )
            {
               operands.push( (int)Math.pow(x,y) );
            }
         }
         else // must be a number
         {
            operands.push( Integer.decode( token ) );
         }
      }
      return operands.pop();
   }//evaluateInfix()



   public static void main(String[] args)
   {
      String[] expressions = {
                               "( 2 + ( 5 * 5 ) )",
                               "( ( 2 + 5 ) * 5 )",
                               "( ( ( 3 + 5 ) * 4 ) - 9 )",
                               "( ( ( 1 + 3 ) ^ 4 ) * ( 2 ^ 3 ) )",
                               "( ( ( 32 * 41 ) / 16 ) + ( 12 * 10 ) )"
                             };

      for (int i = 0; i < expressions.length; i++)
      {
         int result = evaluateInfix( expressions[i] );
         System.out.println( expressions[i] + " ==> " + result );
      }
   }//main()

}